#include <cstdio>
#include <fstream>
#include <string>
#include <complex>
#include "../include/Configuration.h"
#include "../include/sph_subs.h"

using namespace std;

//! \brief C++ implementation of SPH
void sphere() {
	complex<double> arg, s0, tfsas;
	complex<double> **tqspe, **tqsps;
	double **tqse, **tqss;
	double *argi, *args, *gaps;
	double th, ph;
	printf("INFO: making legacy configuration ...\n");
	ScattererConfiguration *conf = ScattererConfiguration::from_dedfb("../../test_data/sphere/DEDFB");
	conf->write_formatted("c_OEDFB");
	conf->write_binary("c_TEDF");
	delete conf;
	printf("INFO: reading binary configuration ...\n");
	ScattererConfiguration *sconf = ScattererConfiguration::from_binary("c_TEDF");
	GeometryConfiguration *gconf = GeometryConfiguration::from_legacy("../../test_data/sphere/DSPH");
	if (sconf->number_of_spheres == gconf->number_of_spheres) {
		int nsph = gconf->number_of_spheres;
		C1 *c1 = new C1;
		for (int i = 0; i < nsph; i++) {
			c1->iog[i] = sconf->iog_vec[i];
			c1->nshl[i] = sconf->nshl_vec[i];
			c1->ros[i] = sconf->radii_of_spheres[i];
		}
		for (int i = 0; i < gconf->l_max; i++) {
			c1->rmi[i][0] = complex<double>(0.0, 0.0);
			c1->rmi[i][1] = complex<double>(0.0, 0.0);
			c1->rei[i][0] = complex<double>(0.0, 0.0);
			c1->rei[i][1] = complex<double>(0.0, 0.0);
		}
		C2 *c2 = new C2;
		argi = new double[1];
		args = new double[1];
		gaps = new double[2];
		FILE *output = fopen("c_OSPH", "w");
		fprintf(output, " READ(IR,*)NSPH,LM,INPOL,NPNT,NPNTTS,ISAM\n");
		fprintf(
				output,
				" %5d%5d%5d%5d%5d%5d\n",
				gconf->number_of_spheres,
				gconf->l_max,
				gconf->in_pol,
				gconf->npnt,
				gconf->npntts,
				gconf->meridional_type
		);
		fprintf(output, " READ(IR,*)TH,THSTP,THLST,THS,THSSTP,THSLST\n");
		fprintf(
				output,
				" %9.3lE %9.3lE %9.3lE %9.3lE %9.3lE %9.3lE\n",
				gconf->in_theta_start,
				gconf->in_theta_step,
				gconf->in_theta_end,
				gconf->sc_theta_start,
				gconf->sc_theta_step,
				gconf->sc_theta_end
		);
		fprintf(output, " READ(IR,*)PH,PHSTP,PHLST,PHS,PHSSTP,PHSLST\n");
		fprintf(
				output,
				" %9.3lE %9.3lE %9.3lE %9.3lE %9.3lE %9.3lE\n",
				gconf->in_phi_start,
				gconf->in_phi_step,
				gconf->in_phi_end,
				gconf->sc_phi_start,
				gconf->sc_phi_step,
				gconf->sc_phi_end
		);
		fprintf(output, " READ(IR,*)JWTM\n");
		fprintf(
				output,
				" %5d\n",
				gconf->jwtm
		);
		fprintf(output, " READ(ITIN)NSPHT\n");
		fprintf(output, " READ(ITIN)(IOG(I),I=1,NSPH)\n");
		fprintf(output, " READ(ITIN)EXDC,WP,XIP,IDFC,NXI\n");
		fprintf(output, " READ(ITIN)(XIV(I),I=1,NXI)\n");
		fprintf(output, " READ(ITIN)NSHL(I),ROS(I)\n");
		fprintf(output, " READ(ITIN)(RCF(I,NS),NS=1,NSH)\n \n");
		double sml = 1.0e-3;
		int nth = 0, nph = 0;
		if (gconf->in_theta_step != 0.0)
			nth = int((gconf->in_theta_end - gconf->in_theta_start) / gconf->in_theta_step + sml);
		nth += 1;
		if (gconf->in_phi_step != 0.0)
			nph = int((gconf->in_phi_end - gconf->in_phi_start) / gconf->in_phi_step + sml);
		nph += 1;
		int nths = 0, nphs = 0;
		double thsca;
		if (gconf->meridional_type > 1) { // ISAM > 1, fixed scattering angle
			nths = 1;
			thsca = gconf->sc_theta_start - gconf->in_theta_start;
		} else { //ISAM <= 1
			if (gconf->in_theta_step != 0.0)
				nths = int((gconf->sc_theta_end - gconf->sc_theta_start) / gconf->sc_theta_step + sml);
			nths += 1;
			if (gconf->meridional_type == 1) { // ISAM = 1
				nphs = 1;
			} else { // ISAM < 1
				if (gconf->sc_phi_step != 0.0)
					nphs = int((gconf->sc_phi_end - gconf->sc_phi_start) / gconf->sc_phi_step + sml);
				nphs += 1;
			}
		}
		int nk = nth * nph;
		int nks = nths * nphs;
		int nkks = nk * nks;
		int th1 = gconf->in_theta_start;
		int ph1 = gconf->in_phi_start;
		int ths1 = gconf->sc_theta_start;
		int phs1 = gconf->sc_phi_start;
		const double half_pi = acos(0.0);
		const double pi = 2.0 * half_pi;
		double gcs = 0.0;
		for (int i116 = 1; i116 <= nsph; i116++) {
			int iogi = c1->iog[i116 - 1];
			if (iogi >= i116) {
				double gcss = pi * c1->ros[i116 - 1] * c1->ros[i116 - 1];
				c1->gcsv[i116 - 1] = gcss;
				int nsh = c1->nshl[i116 - 1];
				for (int j115 = 1; j115 <= nsh; j115++) {
					c1->rc[i116 - 1][j115 - 1] = sconf->rcf[i116 - 1][j115 - 1] * c1->ros[i116 - 1];
				}
			}
			gcs += c1->gcsv[iogi - 1];
		}
		double ****zpv = new double***[gconf->l_max]; //[gconf->l_max][3][2][2]; // Matrix: dim[LM x 3 x 2 x 2]
		for (int zi = 0; zi < gconf->l_max; zi++) {
			zpv[zi] = new double**[3];
			for (int zj = 0; zj < 3; zj++) {
				zpv[zi][zj] = new double*[2];
				for (int zk = 0; zk < 2; zk++) {
					zpv[zi][zj][zk] = new double[2];
				}
			}
		}
		thdps(gconf->l_max, zpv);
		//DEBUG CODE
		/*
		for (int zi = 0; zi < gconf->l_max; zi++) {
			for (int zj = 0; zj < 3; zj++) {
				for (int zk = 0; zk < 2; zk++) {
					if (zpv[zi][zj][zk][0] != 0.0)
						printf(
								"DEBUG: zpv[%d][%d][%d][0] = %.3lE\n",
								zi, zj, zk, zpv[zi][zj][zk][0]
						);
					if (zpv[zi][zj][zk][1] != 0.0)
						printf(
								"DEBUG: zpv[%d][%d][%d][1] = %.3lE\n",
								zi, zj, zk, zpv[zi][zj][zk][1]
						);
				}
			}
		}
		*/
		//END OF DEBUG CODE
		double exri = sqrt(sconf->exdc);
		fprintf(output, "  REFR. INDEX OF EXTERNAL MEDIUM=%15.7lE\n", exri);
		fstream tppoan;
		tppoan.open("c_TPPOAN", ios::binary|ios::out);
		if (tppoan.is_open()) {
			int imode = 10;
			tppoan.write(reinterpret_cast<char *>(&imode), sizeof(int));
			tppoan.write(reinterpret_cast<char *>(&(gconf->meridional_type)), sizeof(int));
			tppoan.write(reinterpret_cast<char *>(&(gconf->in_pol)), sizeof(int));
			tppoan.write(reinterpret_cast<char *>(&(sconf->number_of_scales)), sizeof(int));
			tppoan.write(reinterpret_cast<char *>(&nth), sizeof(int));
			tppoan.write(reinterpret_cast<char *>(&nph), sizeof(int));
			tppoan.write(reinterpret_cast<char *>(&nths), sizeof(int));
			tppoan.write(reinterpret_cast<char *>(&nphs), sizeof(int));

			for (int nsi = 0; nsi < nsph; nsi++)
				tppoan.write(reinterpret_cast<char *>(&(sconf->iog_vec[nsi])), sizeof(int));
			if (gconf->in_pol == 0) fprintf(output, "   LIN\n");
			else fprintf(output, "  CIRC\n");
			fprintf(output, " \n");
			double wn = sconf->wp / 3.0e8;
			double sqsfi = 1.0;
			double vk, vkarg;
			if (sconf->idfc < 0) {
				vk = sconf->xip * wn;
				fprintf(output, "  VK=%15.7lE, XI IS SCALE FACTOR FOR LENGTHS\n", vk);
				fprintf(output, " \n");
			}
			for (int jxi488 = 1; jxi488 <= sconf->number_of_scales; jxi488++) {
				fprintf(output, "========== JXI =%3d ====================\n", jxi488);
				double xi = sconf->scale_vec[jxi488 - 1];
				if (sconf->idfc >= 0) {
					vk = xi * wn;
					vkarg = vk;
					fprintf(output, "  VK=%15.7lE, XI=%15.7lE\n", xi, vk);
				} else { // IDFC < 0
					vkarg = xi * vk;
					sqsfi = 1.0 / (xi * xi);
					fprintf(output, "  XI=%15.7lE\n", xi);
				}
				tppoan.write(reinterpret_cast<char *>(&vk), sizeof(double));
				for (int i132 = 1; i132 <= nsph; i132++) {
					int iogi = sconf->iog_vec[i132 - 1];
					if (iogi != i132) {
						for (int l123 = 1; l123 <= gconf->l_max; l123++) {
							c1->rmi[l123 - 1][i132 - 1] = c1->rmi[l123 - 1][iogi - 1];
							c1->rei[l123 - 1][i132 - 1] = c1->rei[l123 - 1][iogi - 1];
						}
						continue; // i132
					}
					int nsh = sconf->nshl_vec[i132 - 1];
					int ici = (nsh + 1) / 2;
					if (sconf->idfc == 0) {
						for (int ic = 0; ic < ici; ic++)
							c2->dc0[ic] = sconf->dc0_matrix[ic][i132 - 1][0]; // WARNING: IDFC=0 is not tested!
					} else { // IDFC != 0
						if (jxi488 == 1) {
							for (int ic = 0; ic < ici; ic++) {
								c2->dc0[ic] = sconf->dc0_matrix[ic][i132 - 1][jxi488 - 1];
							}
						}
					}
					if (nsh % 2 == 0) c2->dc0[ici] = sconf->exdc;
					int jer = 0;
					int lcalc = 0;
					dme(
							gconf->l_max, i132, gconf->npnt, gconf->npntts, vkarg,
							sconf->exdc, exri, c1, c2, jer, lcalc, arg
					);
					if (jer != 0) {
						fprintf(output, "  STOP IN DME\n");
						fprintf(output, "  AT %1d LCALC=%3d TOO SMALL WITH ARG=%15.7lE+i(%15.7lE)\n", jer, lcalc, arg.real(), arg.imag());
						tppoan.close();
						fclose(output);
						return;
					}
				} // i132
				if (sconf->idfc >= 0 and nsph == 1 and jxi488 == gconf->jwtm) {
					// This is the condition that writes the transition matrix to output.
					int is = 1111;
					fstream ttms;
					ttms.open("c_TTMS", ios::binary | ios::out);
					if (ttms.is_open()) {
						ttms.write(reinterpret_cast<char *>(&is), sizeof(int));
						ttms.write(reinterpret_cast<char *>(&(gconf->l_max)), sizeof(int));
						ttms.write(reinterpret_cast<char *>(&vk), sizeof(double));
						ttms.write(reinterpret_cast<char *>(&exri), sizeof(double));
						for (int lmi = 0; lmi < gconf->l_max; lmi++) {
							complex<double> element1 = -1.0 / c1->rmi[0][lmi];
							complex<double> element2 = -1.0 / c1->rei[0][lmi];
							ttms.write(reinterpret_cast<char *>(&element1), sizeof(complex<double>));
							ttms.write(reinterpret_cast<char *>(&element2), sizeof(complex<double>));
						}
						ttms.write(reinterpret_cast<char *>(&(sconf->radii_of_spheres[0])), sizeof(double));
						ttms.close();
					} else { // Failed to open output file. Should never happen.
						printf("ERROR: could not open TTMS file.\n");
						tppoan.close();
						fclose(output);
					}
				}
				// We are at line 271 of SPH.f
				double cs0 = 0.25 * vk * vk * vk / half_pi;
				sscr0(tfsas, nsph, gconf->l_max, vk, exri, c1);
				printf("DEBUG: TFSAS = (%lE,%lE)\n", tfsas.real(), tfsas.imag());
				double sqk = vk * vk * sconf->exdc;
				aps(zpv, gconf->l_max, nsph, c1, sqk, gaps);
				tqse = new double*[2];
				tqss = new double*[2];
				tqspe = new std::complex<double>*[2];
				tqsps = new std::complex<double>*[2];
				for (int ti = 0; ti < 2; ti++) {
					tqse[ti] = new double[2];
					tqss[ti] = new double[2];
					tqspe[ti] = new std::complex<double>[2];
					tqsps[ti] = new std::complex<double>[2];
				}
				rabas(gconf->in_pol, gconf->l_max, nsph, c1, tqse, tqspe, tqss, tqsps);
				for (int i170 = 1; i170 <= nsph; i170++) {
					if (c1->iog[i170 - 1] >= i170) {
						double albeds = c1->sscs[i170 - 1] / c1->sexs[i170 - 1];
						c1->sqscs[i170 - 1] *= sqsfi;
						c1->sqabs[i170 - 1] *= sqsfi;
						c1->sqexs[i170 - 1] *= sqsfi;
						fprintf(output, "    SPHERE %2d\n", i170);
						if (c1->nshl[i170 - 1] != 1) {
							fprintf(output, "  SIZE=%15.7lE\n", c2->vsz[i170 - 1]);
						} else {
							fprintf(
									output,
									"  SIZE=%15.7lE, REFRACTIVE INDEX=(%15.7lE,%15.7lE)\n",
									c2->vsz[i170 -1],
									c2->vkt[i170 - 1].real(),
									c2->vkt[i170 - 1].imag()
							);
						}
						fprintf(output, " ----- SCS ----- ABS ----- EXS ----- ALBEDS --\n");
						fprintf(
								output,
								" %14.7lE%15.7lE%15.7lE%15.7lE\n",
								c1->sscs[i170 - 1], c1->sabs[i170 - 1],
								c1->sexs[i170 - 1], albeds
						);
						fprintf(output, " ---- SCS/GS -- ABS/GS -- EXS/GS ---\n");
						fprintf(
								output,
								" %14.7lE%15.7lE%15.7lE\n",
								c1->sqscs[i170 - 1], c1->sqabs[i170 - 1],
								c1->sqexs[i170 - 1]
						);
						fprintf(output, "  FSAS=%15.7lE%15.7lE\n", c1->fsas[i170 - 1].real(), c1->fsas[i170 - 1].imag());
						double csch = 2.0 * vk * sqsfi / c1->gcsv[i170 -1];
						s0 = c1->fsas[i170 - 1] * exri;
						double qschu = csch * s0.imag();
						double pschu = csch * s0.real();
						double s0mag = cs0 * sqrt((s0.real() + s0.imag()) * (s0.real() - s0.imag()));
						fprintf(
								output,
								"  QSCHU=%15.7lE, PSCHU=%15.7lE, S0MAG=%15.7lE\n",
								qschu, pschu, s0mag
						);
						double rapr = c1->sexs[i170 - 1] - gaps[i170 - 1];
						double cosav = gaps[i170 - 1] / c1->sscs[i170 - 1];
						fprintf(output, "  COSAV=%15.7lE, RAPRS=%15.7lE\n", cosav, rapr);
						fprintf(output, "  IPO=%2d, TQEk=%15.7lE, TQSk=%15.7lE\n", 1, tqse[0][i170 - 1], tqss[0][i170 - 1]);
						fprintf(output, "  IPO=%2d, TQEk=%15.7lE, TQSk=%15.7lE\n", 2, tqse[1][i170 - 1], tqss[1][i170 - 1]);
						tppoan.write(reinterpret_cast<char *>(&(tqse[0][i170 - 1])), sizeof(double));
						tppoan.write(reinterpret_cast<char *>(&(tqss[0][i170 - 1])), sizeof(double));
						double val = tqspe[0][i170 - 1].real();
						tppoan.write(reinterpret_cast<char *>(&val), sizeof(double));
						val = tqspe[0][i170 - 1].imag();
						tppoan.write(reinterpret_cast<char *>(&val), sizeof(double));
						val = tqsps[0][i170 - 1].real();
						tppoan.write(reinterpret_cast<char *>(&val), sizeof(double));
						val = tqsps[0][i170 - 1].imag();
						tppoan.write(reinterpret_cast<char *>(&val), sizeof(double));
						tppoan.write(reinterpret_cast<char *>(&(tqse[1][i170 - 1])), sizeof(double));
						tppoan.write(reinterpret_cast<char *>(&(tqss[1][i170 - 1])), sizeof(double));
						val = tqspe[1][i170 - 1].real();
						tppoan.write(reinterpret_cast<char *>(&val), sizeof(double));
						val = tqspe[1][i170 - 1].imag();
						tppoan.write(reinterpret_cast<char *>(&val), sizeof(double));
						val = tqsps[1][i170 - 1].real();
						tppoan.write(reinterpret_cast<char *>(&val), sizeof(double));
						val = tqsps[1][i170 - 1].imag();
						tppoan.write(reinterpret_cast<char *>(&val), sizeof(double));
					} // End if iog[i170 - 1] >= i170
				} // i170 loop
				if (nsph != 1) {
					fprintf(output, "  FSAT=(%15.7lE,%15.7lE)\n", tfsas.real(), tfsas.imag());
					double csch = 2.0 * vk * sqsfi / gcs;
					s0 = tfsas * exri;
					double qschu = csch * s0.imag();
					double pschu = csch * s0.real();
					double s0mag = cs0 * sqrt((s0.real() + s0.imag()) * (s0.real() - s0.imag()));
					fprintf(
							output,
							"  QSCHU=%15.7lE, PSCHU=%15.7lE, S0MAG=%15.7lE\n",
							qschu, pschu, s0mag
					);
				}
				th = gconf->in_theta_start;
				int isq, ibf;
				double cost, sint, cosp, sinp;
				double costs, sints, cosps, sinps;
				double scan;
				double *duk = new double[3];
				double *u = new double[3];
				double *us = new double[3];
				double *un = new double[3];
				double *uns = new double[3];
				double *up = new double[3];
				double *ups = new double[3];
				double *upmp = new double[3];
				double *upsmp = new double[3];
				double *unmp = new double[3];
				double *unsmp = new double[3];
				double frx, fry, frz;
				double cfmp, cfsp, sfmp, sfsp;
				int jw;
				for (int jth486 = 1; jth486 <= nth; jth486++) { // OpenMP parallelizable section
					ph = gconf->in_phi_start;
					for (int jph484 = 1; jph484 <= nph; jph484++) {
						bool goto182 = (nk == 1) && (jxi488 > 1);
						if (!goto182) {
							upvmp(th, ph, 0, cost, sint, cosp, sinp, u, upmp, unmp);
						}
						if (gconf->meridional_type >= 0) {
							wmamp(0, cost, sint, cosp, sinp, gconf->in_pol, gconf->l_max, 0, nsph, argi, u, upmp, unmp, c1);
							for (int i183 = 0; i183 < nsph; i183++) {
								double rapr = c1->sexs[i183] - gaps[i183];
								frx = rapr * u[0];
								fry = rapr * u[1];
								frx = rapr * u[2];
							}
							jw = 1;
						}
						double thsl = ths1;
						double phsph = 0.0;
						for (int jths482 = 1; jths482 <= nths; jths482++) {
							double ths = thsl;
							int icspnv = 0;
							if (gconf->meridional_type > 1) ths = th + thsca;
							if (gconf->meridional_type >= 1) {
								phsph = 0.0;
								if ((ths < 0.0) || (ths > 180.0)) phsph = 180.0;
								if (ths < 0.0) ths *= -1.0;
								if (ths > 180.0) ths = 360.0 - ths;
								if (phsph != 0.0) icspnv = 1;
							}
							double phs = phs1;
							for (int jphs480 = 1; jphs480 <= nphs; jphs480++) {
								if (gconf->meridional_type >= 1) {
									phs = ph + phsph;
									if (phs >= 360.0) phs -= 360.0;
								}
								bool goto190 = ((nks == 1) && (jxi488 > 1)) || (jth486 > 1) || (jph484 > 1);
								if (!goto190) {
									upvmp(ths, phs, icspnv, costs, sints, cosps, sinps, us, upsmp, unsmp);
									if (gconf->meridional_type >= 0) {
										wmamp(2, costs, sints, cosps, sinps, gconf->in_pol, gconf->l_max, 0, nsph, args, us, upsmp, unsmp, c1);
									}
								}
								if (nkks != 0 || jxi488 == 1) {
									upvsp(u, upmp, unmp, us, upsmp, unsmp, up, un, ups, uns, duk, isq, ibf, scan, cfmp, sfmp, cfsp, sfsp);
									if (gconf->meridional_type < 0) {
										wmasp(
												cost, sint, cosp, sinp, costs, sints, cosps, sinps,
												u, up, un, us, ups, uns, isq, ibf, gconf->in_pol,
												gconf->l_max, 0, nsph, argi, args, c1
										);
									}
									for (int i193 = 1; i193 <= 3; i193++) {
										un[i193 - 1] = unmp[i193 - 1];
										uns[i193 - 1] = unsmp[i193 - 1];
									}
								}
								if (gconf->meridional_type < 0) jw = 1;
								tppoan.write(reinterpret_cast<char *>(&th), sizeof(double));
								tppoan.write(reinterpret_cast<char *>(&ph), sizeof(double));
								tppoan.write(reinterpret_cast<char *>(&ths), sizeof(double));
								tppoan.write(reinterpret_cast<char *>(&phs), sizeof(double));
								tppoan.write(reinterpret_cast<char *>(&scan), sizeof(double));
								if (jw != 0) {
									jw = 0;
									tppoan.write(reinterpret_cast<char *>(&(u[0])), sizeof(double));
									tppoan.write(reinterpret_cast<char *>(&(u[1])), sizeof(double));
									tppoan.write(reinterpret_cast<char *>(&(u[2])), sizeof(double));
								}
								fprintf(
										output,
										"********** JTH =%3d, JPH =%3d, JTHS =%3d, JPHS =%3d ********************\n",
										jth486, jph484, jths482, jphs480
								);
								fprintf(
										output,
										"  TIDG=%10.3lE, PIDG=%10.3lE, TSDG=%10.3lE, PSDG=%10.3lE\n",
										th, ph, ths, phs
								);
								fprintf(output, "  SCAND=%10.3lE\n", scan);
								fprintf(output, "  CFMP=%15.7lE, SFMP=%15.7lE\n", cfmp, sfmp);
								fprintf(output, "  CFSP=%15.7lE, SFSP=%15.7lE\n", cfsp, sfsp);
								if (gconf->meridional_type >= 0) {
									fprintf(output, "  UNI=(%12.5lE,%12.5lE,%12.5lE)\n", un[0], un[1], un[2]);
									fprintf(output, "  UNS=(%12.5lE,%12.5lE,%12.5lE)\n", uns[0], uns[1], uns[2]);
								} else {
									fprintf(output, "  UN=(%12.5lE,%12.5lE,%12.5lE)\n", un[0], un[1], un[2]);
								}

								if (gconf->meridional_type < 1) phs += gconf->sc_phi_step;
							} // jphs480 loop
							if (gconf->meridional_type <= 1) thsl += gconf->sc_theta_step;
						} // jths482 loop
						ph += gconf->in_phi_step;
					} // jph484 loop on elevation
					th += gconf->in_theta_step;
				} // jth486 loop on azimuth
			} //jxi488 loop on scales
			tppoan.close();
		} else { // In case TPPOAN could not be opened. Should never happen.
			printf("ERROR: failed to open TPPOAN file.\n");
		}
		fclose(output);
	} else { // NSPH mismatch between geometry and scatterer configurations.
		throw UnrecognizedConfigurationException(
				"Inconsistent geometry and scatterer configurations."
		);
	}
}
